﻿using System;
using Eneter.Messaging.DataProcessing.Serializing;
using Eneter.Messaging.EndPoints.TypedMessages;
using Eneter.Messaging.MessagingSystems.MessagingSystemBase;
using Eneter.Messaging.MessagingSystems.WebSocketMessagingSystem;

namespace PiService
{
    public class PiRequestMessage
    {
        public double CalculationStep { get; set; }
    }

    public class PiResponseMessage
    {
        public double Result { get; set; }
    }

    class Program
    {
        static void Main(string[] args)
        {
            // Use JSON serializer which can be used by JavaScript too.
            ISerializer aJsonSerialzier = new DataContractJsonStringSerializer();

            // Factory to create message receivers.
            IDuplexTypedMessagesFactory aReceiverFactory = new DuplexTypedMessagesFactory(aJsonSerialzier);

            // Create message receiver.
            IDuplexTypedMessageReceiver<PiResponseMessage, PiRequestMessage> aReceiver
                = aReceiverFactory.CreateDuplexTypedMessageReceiver<PiResponseMessage, PiRequestMessage>();

            // Subscribe to process incoming messages.
            aReceiver.MessageReceived += OnMessageReceived;

            // Subscribe to observe client connections/disconnections.
            aReceiver.ResponseReceiverConnected += OnClientConnected;
            aReceiver.ResponseReceiverDisconnected += OnClientDisconnected;

            // Use WebSocket for the communication.
            IMessagingSystemFactory aMessaging = new WebSocketMessagingSystemFactory();

            // Attach input channel and be able to receive request messages
            // and send back response messages.
            IDuplexInputChannel anInputChennel = aMessaging.CreateDuplexInputChannel("ws://127.0.0.1:8091/PiCalculator/");
            aReceiver.AttachDuplexInputChannel(anInputChennel);

            Console.WriteLine("Pi service is running. Press ENTER to stop.");
            Console.ReadLine();

            // Detach input channel and stop listening.
            // Note: it releases the listening thread.
            aReceiver.DetachDuplexInputChannel();
        }

        
        // Method which is invoked when the request message is received.
        static void OnMessageReceived(object sender, TypedRequestReceivedEventArgs<PiRequestMessage> e)
        {
            // Get the receiver instance.
            IDuplexTypedMessageReceiver<PiResponseMessage, PiRequestMessage> aReceiver
                = (IDuplexTypedMessageReceiver<PiResponseMessage, PiRequestMessage>)sender;

            PiResponseMessage aResponseMessage = new PiResponseMessage();

            // Calculate PI and send back multiple response messages.
            double aResult = 0.0;
            double aDx = e.RequestMessage.CalculationStep;
            for (double x = -1.0; x < 1.0; x += aDx)
            {
                aResult += 2 * Math.Sqrt(1 - x * x) * aDx;

                // Send back the current result.
                // Note: ResponseReceiverId identifies the client who sent the request.
                aResponseMessage.Result = aResult;
                aReceiver.SendResponseMessage(e.ResponseReceiverId, aResponseMessage);
            }
        }

        private static void OnClientConnected(object sender, ResponseReceiverEventArgs e)
        {
            Console.WriteLine("Client connected: " + e.ResponseReceiverId);
        }

        private static void OnClientDisconnected(object sender, ResponseReceiverEventArgs e)
        {
            Console.WriteLine("Client disconnected: " + e.ResponseReceiverId);
        }
        
    }
}
